import argparse
import numpy as np

import torch
import os


parser = argparse.ArgumentParser('ODE demo')
parser.add_argument('--method', type=str, choices=['dopri5', 'adams'], default='dopri5')
parser.add_argument('--seq_len', type=int, default=20)
parser.add_argument('--gpu', type=int, default=0)
parser.add_argument("--dataset", default="walker2d")
parser.add_argument('--adjoint', action='store_true')
parser.add_argument('--consider_noise', action='store_true')
parser.add_argument('--hard_truncate', action='store_true')
parser.add_argument('--num_var', type=int, default=1)
parser.add_argument('--use_dcbf', action='store_true')
args = parser.parse_args()

from train_walker import ODEFunc
# from walker2d_data import load_dataset
import walker2d_data 
import cheetah_data

if args.adjoint:
    from torchdiffeq import odeint_adjoint as odeint
else:
    from torchdiffeq import odeint

os.environ['CUDA_VISIBLE_DEVICES'] = '1'
device = torch.device("cuda:" + str(args.gpu) if torch.cuda.is_available() else "cpu")
device = "cpu"


batch_id = 0
time_delta = 0.1

# trainloader, testloader, in_features = load_dataset(seq_len=args.seq_len)


if args.dataset == "walker2d":
    trainloader, testloader, in_features = walker2d_data.load_dataset(
        seq_len=args.seq_len
    )
    low = torch.from_numpy(
        np.array(
            [
                0.8306609392166138,
                -0.629609227180481,
                -1.1450262069702148,
                -1.2276967763900757,
                -1.2756223678588867,
                -0.9984434843063354,
                -1.2026487588882446,
                -1.4123592376708984,
                -2.2933568954467773,
                -3.8838634490966797,
                -10.0,
                -10.0,
                -10.0,
                -10.0,
                -10.0,
                -10.0,
                -10.0,
            ]
        )
    ).to(device)

    high = torch.from_numpy(
        np.array(
            [
                1.6924704313278198,
                0.9378023743629456,
                0.1535143405199051,
                0.16713030636310577,
                1.2674068212509155,
                0.2375510334968567,
                0.18477311730384827,
                1.2249037027359009,
                5.372719764709473,
                3.470100164413452,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
                10.0,
            ]
        )
    ).to(device)
elif args.dataset == "halfcheetah":
    trainloader, testloader, in_features = cheetah_data.load_dataset(
        seq_len=args.seq_len
    )
    low = torch.from_numpy(
        np.array(
            [
                -0.2936522662639618,
                -0.29987457394599915,
                -0.5890289545059204,
                -0.7020226120948792,
                -0.518144965171814,
                -0.7397885918617249,
                -0.841916024684906,
                -0.6131184697151184,
                -2.262007713317871,
                -2.206634759902954,
                -4.519010066986084,
                -17.817516326904297,
                -17.485055923461914,
                -21.776256561279297,
                -18.28421401977539,
                -20.88571548461914,
                -19.104284286499023,
            ]
        )
    ).to(device)

    high = torch.from_numpy(
        np.array(
            [
                0.1298145353794098,
                0.4833938181400299,
                0.6584464311599731,
                0.6455742120742798,
                0.8780248761177063,
                0.7083053588867188,
                0.7404807209968567,
                0.5502343773841858,
                4.839046955108643,
                2.9650542736053467,
                4.737936019897461,
                16.746498107910156,
                22.432693481445312,
                18.447025299072266,
                22.226903915405273,
                22.869461059570312,
                15.584585189819336,
            ]
        )
    ).to(device)
else:
    raise ValueError(f"Unknown dataset '{args.dataset}'")

func = ODEFunc(in_features).to(device)
func_trunc = ODEFunc(in_features).to(device)


if args.dataset == "walker2d":
    func.load_state_dict(torch.load(f"ckpt/walker2d_ep070.pth"))
    func.eval()
    func_trunc.load_state_dict(torch.load(f"ckpt_tunc/walker2d_ep020.pth"))
    func_trunc.eval()
else:
    func.load_state_dict(torch.load(f"ckpt/halfcheetah_ep140.pth"))
    func.eval()
    func_trunc.load_state_dict(torch.load(f"ckpt_tunc/halfcheetah_ep140.pth"))
    func_trunc.eval()
total = len(testloader) - 1
with torch.no_grad():
    test_losses = []
    test_losses_ode = []
    test_losses_trunc = []
    # for batch_y0, batch_t, batch_y in iter(testloader):
    for batch_y0, batch_y in iter(testloader):
        batch_y0 = batch_y0.to(device)
        
        ################################################invariance
        batch_y0 = batch_y0[:,0,:]
        batch_t = torch.linspace(
            time_delta, args.seq_len * time_delta, args.seq_len).to(device)
        batch_y = batch_y.to(device)
        
        func.use_dcbf = True
        batch_t = torch.tensor([0., 0.1]).to(device)
        seq = batch_y.shape[1]
        nbat = batch_y.shape[0]
        pred_y = []
        for j in range(nbat):
            y0 = batch_y0[j,:]
            pred_yj = []
            for i in range(seq):
                print('batch_id: ', batch_id, '/', total, 'seq: ', j, 'point: ', i)
                pred = odeint(func, y0, batch_t)
                pred_yj.append(pred[-1,:].unsqueeze(0))
                y0 = pred[-1,:]
            pred_yj = torch.cat(pred_yj, dim = 0)
            pred_y.append(pred_yj.unsqueeze(0))
        pred_y = torch.cat(pred_y, dim=0)
       
        loss = torch.mean(torch.abs(pred_y - batch_y)).detach().cpu().item()
        test_losses.append(loss)
        print(f"Invarince Test_loss ={np.mean(test_losses):0.4g}")

        ##################################################neural ODE
        batch_t = torch.linspace(
            time_delta, args.seq_len * time_delta, args.seq_len).to(device)
        func.use_dcbf = False

        pred_y = odeint(func, batch_y0, batch_t)
        pred_y = torch.transpose(pred_y, 0, 1)
       
        loss = torch.mean(torch.abs(pred_y - batch_y)).detach().cpu().item()
        test_losses_ode.append(loss)
        print(f"neural ODE Test_loss ={np.mean(test_losses_ode):0.4g}")

        ##################################################hard trunc
        func_trunc.use_dcbf = False
        pred_y = odeint(func_trunc, batch_y0, batch_t)
        pred_y = torch.transpose(pred_y, 0, 1)

        trunc_y = torch.clip(pred_y, low, high)
        loss = torch.mean(torch.abs(trunc_y - batch_y)).detach().cpu().item()
        test_losses_trunc.append(loss)
        print(f"hard trunc Test_loss ={np.mean(test_losses_trunc):0.4g}")

        batch_id = batch_id + 1

    print(f"Invarince Test_loss ={np.mean(test_losses):0.4g}")
    print(f"Invarince Test_loss_std ={np.std(test_losses):0.4g}")
    print(f"neural ODE Test_loss ={np.mean(test_losses_ode):0.4g}")
    print(f"neural ODE Test_loss_std ={np.std(test_losses_ode):0.4g}")
    print(f"hard trunc Test_loss ={np.mean(test_losses_trunc):0.4g}")
    print(f"hard trunc Test_loss_std ={np.std(test_losses_trunc):0.4g}")